"""
*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
"""

from __future__ import print_function

import logging
import sys

_MESSAGE_CATEGORY = 2


class ErrorMessage:  #pylint: disable=no-init,too-few-public-methods
    """Enum for Error Messages."""
    UNEXPECTED_ERROR = 0
    NUM_PROC_UNSUPPORTED = 1
    RANKGEN_PARAM_UNSUPPORTED = 2
    HW_SPEC_ERROR = 3
    DDL_OPTIONS_SET = 4
    RANK_ARG_FILE = 5
    RANK_FILE_HOST = 6
    TOPO_NO_DDLOPTS = 7
    MODE_NO_DDLOPTS = 8
    DDLARGS_NO_DDLOPTS = 9
    RANK_FILE_ERROR = 10
    SW_ERROR = 11
    HOST_CFG = 12
    DIFF_HOST_CFG = 13
    NO_RANKFILE = 14
    NO_MPIRUN = 15
    MPI_OPS = 16
    NO_MPITOOL = 17
    SSH_ERROR = 18
    NO_AUTO_RANKFILE = 19
    INVALID_TOPO = 20
    DIFF_CORES = 21
    NO_EXECUTABLE = 22
    HOSTFILE_AND_HOSTLIST = 23
    HOSTFILE_AND_RANKFILE = 24
    NO_HOSTFILE = 25
    HOSTFILE_COMMA_SPACE = 26
    INVALID_RANKGEN_ARG = 27
    SKIPCHECKS_SET = 28
    INFINIBAND_NOT_SETUP = 29
    INVALID_PORT = 30
    NO_REMOTE_CONDA_ENV = 31
    NO_LOCAL_CONDA_ENV = 32
    NO_REMOTE_CONDA_SETUP_SCRIPT = 33
    NO_LOCAL_CONDA_SETUP_SCRIPT = 34
    NO_LOCAL_CONDA_INSTALL = 35
    INVALID_MPI_OPS = 36
    LSF_AND_HOSTS = 37
    NO_LSF_ENV_VAR = 38
    NODE_COUNT_CHECK_SKIP = 39
    NO_LSF_JOB_ENV_VAR = 40

DDL_ERRORS = {
    ErrorMessage.UNEXPECTED_ERROR:
    "Unexpected Error: %s",
    ErrorMessage.NUM_PROC_UNSUPPORTED:
    "Specifying the number of processes with %s has been deprecated and is no longer supported.",
    ErrorMessage.RANKGEN_PARAM_UNSUPPORTED:
    "%s has been deprecated and is no longer supported.",
    ErrorMessage.HW_SPEC_ERROR:
    "An error was encountered trying to determine local hardware specs, "
    "defaulting to use a single mpi instance.",
    ErrorMessage.DDL_OPTIONS_SET:
    "The environment variable DDL_OPTIONS is set, but is being overwritten by ddlrun.",
    ErrorMessage.RANK_ARG_FILE:
    "Rankgen args were provided, with a rankfile.",
    ErrorMessage.RANK_FILE_HOST:
    "A rankfile and a host list cannot be used at the same time.",
    ErrorMessage.TOPO_NO_DDLOPTS:
    "Topology is being passed in with the --no_ddloptions flag.",
    ErrorMessage.MODE_NO_DDLOPTS:
    "Mode is being passed in with the --no_ddloptions flag.",
    ErrorMessage.DDLARGS_NO_DDLOPTS:
    "ddlargs are being passed in with the --no_ddloptions flag.",
    ErrorMessage.RANK_FILE_ERROR:
    "An error was encountered trying to generate the rankfile.",
    ErrorMessage.SW_ERROR:
    "An error was encountered verifying the software exists on host '%s'.\n"
    "Continuing to test other host software.",
    ErrorMessage.HOST_CFG:
    "An error was encountered verifying the host config on host '%s'.\n"
    "Continuing to test host configurations.",
    ErrorMessage.DIFF_HOST_CFG:
    "All machine configurations are not the same. "
    "Please verify that the correct hosts are being used.",
    ErrorMessage.NO_RANKFILE:
    "Cannot find rankfile %s.",
    ErrorMessage.NO_MPIRUN:
    "Cannot find 'mpirun' executable. Do you need to run an activate script?",
    ErrorMessage.MPI_OPS:
    "An error was encountered trying to use the specified MPI options.",
    ErrorMessage.NO_MPITOOL:
    "Not performing connection tests. Cannot find 'mpitool' executable. "
    "This could be because you are using a version of mpi that does not ship with mpitool.",
    ErrorMessage.SSH_ERROR:
    "An error was encountered trying to ssh to host '%s'.\nContinuing to test host connectivity.",
    ErrorMessage.NO_AUTO_RANKFILE:
    "%s does not appear to be an auto-generated rankfile.\n%s",
    ErrorMessage.INVALID_TOPO:
    "%s has invalid topology tags. #aisles X #racks must be less than "
    "or equal to the number of hosts.",
    ErrorMessage.DIFF_CORES:
    "Hosts have different numbers of cores. The minimum core count is being used. "
    "This may lead to sub-optimal performance.",
    ErrorMessage.NO_EXECUTABLE:
    "The program %s does not exist or is not executable on host '%s'.",
    ErrorMessage.HOSTFILE_AND_HOSTLIST:
    "A host file and host list cannot be used at the same time.",
    ErrorMessage.HOSTFILE_AND_RANKFILE:
    "A host file and rankfile cannot be used at the same time.",
    ErrorMessage.NO_HOSTFILE:
    "Specified hostfile %s does not exist, or is not accessible.",
    ErrorMessage.HOSTFILE_COMMA_SPACE:
    "Hostfile %s contains lines with uncommented commas and/or spaces. "
    "Only one host name is allowed per line.",
    ErrorMessage.INVALID_RANKGEN_ARG:
    "The value %s is not a valid option for %s. "
    "All rankgen/topology arguments must be positive integers.",
    ErrorMessage.SKIPCHECKS_SET:
    "The SKIPCHECKS flag is set. This removes the ability for 'ddlrun' to detect "
    "remote hardware configurations and can cause errors in non-homogenous cluster "
    "environments - i.e., machines with varying core or GPU counts.",
    ErrorMessage.INFINIBAND_NOT_SETUP:
    "InfiniBand is not properly configured on host '%s'. Use '--mpiarg -pami_noib' flag for "
    "single host runs or '--tcp' for multi-host TCP communication.",
    ErrorMessage.INVALID_PORT:
    "An invalid port was specified: '%s'.",
    ErrorMessage.NO_REMOTE_CONDA_ENV:
    "A conda environment with the name '%s' does not exist on host '%s'.\n"
    "Continuing to test other host software.",
    ErrorMessage.NO_LOCAL_CONDA_ENV:
    "There is no active local conda environment.\n"
    "Activate a conda environment with 'conda activate <env_name>' before using 'ddlrun'.",
    ErrorMessage.NO_REMOTE_CONDA_SETUP_SCRIPT:
    "A valid conda setup script cannot be found on host '%s'.\n"
    "Expected setup script: %s",
    ErrorMessage.NO_LOCAL_CONDA_SETUP_SCRIPT:
    "A valid conda setup script cannot be found on the local machine.\n"
    "Expected setup script path: $CONDA_EXE/../../etc/profile.d/conda.sh",
    ErrorMessage.NO_LOCAL_CONDA_INSTALL:
    "A valid conda installation cannot be found on the local machine. 'CONDA_EXE' is not set",
    ErrorMessage.INVALID_MPI_OPS:
    "An error was encountered trying to connect via mpi to host '%s'.",
    ErrorMessage.NO_LSF_ENV_VAR:
    "Trying to parse LSF hosts, but the 'LSB_MCPU_HOSTS' environment variable is not defined.",
    ErrorMessage.LSF_AND_HOSTS:
    "It appears that you are running in LSF, but are manually specifying hosts with a rankfile, "
    "hostfile or host list.\n"
    "'ddlrun' can automatically retrieve LSF hosts if hosts are not specified.",
    ErrorMessage.NODE_COUNT_CHECK_SKIP:
    "Running on more than %d nodes. Skipping checks for performance.",
    ErrorMessage.NO_LSF_JOB_ENV_VAR:
    "Trying to parse LSF hosts, but the 'LSB_JOBID' environment variable is not defined."
}


class DDLError(Exception):
    """Base class for exceptions in this module."""


class GenericDDLError(DDLError):
    """Generic DDL Exception.

    Attributes:
        error: ddl ErrorMessage object
        additional_args: additional arguments for the error object
    """

    def __init__(self, error, *additional_args):
        DDLError.__init__(self)
        self.error = error
        self.additional_args = additional_args


def log_error(error, *additional_args, **kwargs):
    """Log an ERROR level message."""
    logger = logging.getLogger("ddlrun")
    logger.error("[%s DDL-%d-%d] %s", "ERROR", _MESSAGE_CATEGORY, error,
                 DDL_ERRORS[error] % (additional_args), **kwargs)
    if len(logger.handlers) > 1:
        eprint("Please see %s for detailed log." %
               (logger.handlers[1].baseFilename))


def log_warning(error, *additional_args, **kwargs):
    """Log a WARNING level message."""
    logger = logging.getLogger("ddlrun")
    logger.warning("[%s DDL-%d-%d] %s", "WARN", _MESSAGE_CATEGORY, error,
                   DDL_ERRORS[error] % (additional_args), **kwargs)
    eprint(
        "Please see %s for detailed log." % (logger.handlers[1].baseFilename))


def log_info(error, *additional_args, **kwargs):
    """Log an INFO level message."""
    logger = logging.getLogger("ddlrun")
    if error.strip():
        logger.info(error, *additional_args, **kwargs)


def eprint(*args, **kwargs):
    """Print on stderr."""
    print(*args, file=sys.stderr, **kwargs)


def make_logger(args):
    """Make the file and screen logger."""
    logger = logging.getLogger("ddlrun")
    logger.setLevel(logging.DEBUG)

    display_handler = logging.StreamHandler()
    display_handler.setLevel(logging.INFO)
    file_handler = logging.FileHandler(args.logfile)
    file_handler.setLevel(logging.DEBUG)

    log_format = '%(message)s'
    display_format = logging.Formatter(log_format)
    display_handler.setFormatter(display_format)
    file_format = logging.Formatter('%(asctime)s : ' + log_format)
    file_handler.setFormatter(file_format)

    logger.addHandler(display_handler)
    logger.addHandler(file_handler)


def set_logging_level(args):
    """Update the verbosity of the screen logger."""
    logger = logging.getLogger("ddlrun")
    logger.handlers[0].setLevel(logging.INFO if args.verbose else logging.
                                WARNING)
